
from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x4b98f4C73fadA8E9C3f573502713FCebfA5a98AC",
    "sd": "0xD0E292bAfBB1F691fEAFa7ddE1050EAEa923b2e3"
}

ABI = """[
    {"name": "CraftingGenesRerolled", "type": "event", "inputs": [{"name": "beforeHero", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "afterHero", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "craftingGenesRerolled", "type": "bool", "indexed": false, "internalType": "bool"}], "anonymous": false},
    {"name": "HeroRerollStarted", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "HeroRerolled", "type": "event", "inputs": [{"name": "heroId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "heroBefore", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "heroAfter", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}], "anonymous": false},
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "StatUp", "type": "event", "inputs": [{"name": "player", "type": "address", "indexed": true, "internalType": "address"}, {"name": "heroId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "stat", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "increase", "type": "uint8", "indexed": false, "internalType": "uint8"}, {"name": "updateType", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "TaintedGenesRerolled", "type": "event", "inputs": [{"name": "beforeHero", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "afterHero", "type": "tuple", "internalType": "struct Hero", "indexed": false, "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "geneIndex", "type": "uint8[]", "indexed": false, "internalType": "uint8[]"}, {"name": "rankType", "type": "uint8[]", "indexed": false, "internalType": "uint8[]"}], "anonymous": false},
    {"name": "cancelReroll", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "endRoll", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "heroToRerollData", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "beforeHero", "type": "tuple", "internalType": "struct Hero", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "summoningInfo", "type": "tuple", "internalType": "struct SummoningInfo", "components": [{"name": "summonedTime", "type": "uint256", "internalType": "uint256"}, {"name": "nextSummonTime", "type": "uint256", "internalType": "uint256"}, {"name": "summonerId", "type": "uint256", "internalType": "uint256"}, {"name": "assistantId", "type": "uint256", "internalType": "uint256"}, {"name": "summons", "type": "uint32", "internalType": "uint32"}, {"name": "maxSummons", "type": "uint32", "internalType": "uint32"}]}, {"name": "info", "type": "tuple", "internalType": "struct HeroInfo", "components": [{"name": "statGenes", "type": "uint256", "internalType": "uint256"}, {"name": "visualGenes", "type": "uint256", "internalType": "uint256"}, {"name": "rarity", "type": "uint8", "internalType": "enum Rarity"}, {"name": "shiny", "type": "bool", "internalType": "bool"}, {"name": "generation", "type": "uint16", "internalType": "uint16"}, {"name": "firstName", "type": "uint32", "internalType": "uint32"}, {"name": "lastName", "type": "uint32", "internalType": "uint32"}, {"name": "shinyStyle", "type": "uint8", "internalType": "uint8"}, {"name": "class", "type": "uint8", "internalType": "uint8"}, {"name": "subClass", "type": "uint8", "internalType": "uint8"}]}, {"name": "state", "type": "tuple", "internalType": "struct HeroState", "components": [{"name": "staminaFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "hpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "mpFullAt", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint16", "internalType": "uint16"}, {"name": "xp", "type": "uint64", "internalType": "uint64"}, {"name": "currentQuest", "type": "address", "internalType": "address"}, {"name": "sp", "type": "uint8", "internalType": "uint8"}, {"name": "status", "type": "uint8", "internalType": "enum HeroStatus"}]}, {"name": "stats", "type": "tuple", "internalType": "struct HeroStats", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hp", "type": "uint16", "internalType": "uint16"}, {"name": "mp", "type": "uint16", "internalType": "uint16"}, {"name": "stamina", "type": "uint16", "internalType": "uint16"}]}, {"name": "primaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "secondaryStatGrowth", "type": "tuple", "internalType": "struct HeroStatGrowth", "components": [{"name": "strength", "type": "uint16", "internalType": "uint16"}, {"name": "intelligence", "type": "uint16", "internalType": "uint16"}, {"name": "wisdom", "type": "uint16", "internalType": "uint16"}, {"name": "luck", "type": "uint16", "internalType": "uint16"}, {"name": "agility", "type": "uint16", "internalType": "uint16"}, {"name": "vitality", "type": "uint16", "internalType": "uint16"}, {"name": "endurance", "type": "uint16", "internalType": "uint16"}, {"name": "dexterity", "type": "uint16", "internalType": "uint16"}, {"name": "hpSm", "type": "uint16", "internalType": "uint16"}, {"name": "hpRg", "type": "uint16", "internalType": "uint16"}, {"name": "hpLg", "type": "uint16", "internalType": "uint16"}, {"name": "mpSm", "type": "uint16", "internalType": "uint16"}, {"name": "mpRg", "type": "uint16", "internalType": "uint16"}, {"name": "mpLg", "type": "uint16", "internalType": "uint16"}]}, {"name": "professions", "type": "tuple", "internalType": "struct HeroProfessions", "components": [{"name": "mining", "type": "uint16", "internalType": "uint16"}, {"name": "gardening", "type": "uint16", "internalType": "uint16"}, {"name": "foraging", "type": "uint16", "internalType": "uint16"}, {"name": "fishing", "type": "uint16", "internalType": "uint16"}]}]}, {"name": "seedblock", "type": "uint256", "internalType": "uint256"}, {"name": "finishTime", "type": "uint256", "internalType": "uint256"}, {"name": "rerollStarted", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "multiEndRoll", "type": "function", "inputs": [{"name": "_heroIds", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "multiStartRoll", "type": "function", "inputs": [{"name": "_heroIds", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "startRoll", "type": "function", "inputs": [{"name": "_heroId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class GeneReroll(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def cancel_reroll(self, cred:Credentials, _hero_id:uint256) -> TxReceipt:
        tx = self.contract.functions.cancelReroll(_hero_id)
        return self.send_transaction(tx, cred)

    def end_roll(self, cred:Credentials, _hero_id:uint256) -> TxReceipt:
        tx = self.contract.functions.endRoll(_hero_id)
        return self.send_transaction(tx, cred)

    def hero_to_reroll_data(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[tuple, uint256, uint256, bool]:
        return self.contract.functions.heroToRerollData(a).call(block_identifier=block_identifier)

    def multi_end_roll(self, cred:Credentials, _hero_ids:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.multiEndRoll(_hero_ids)
        return self.send_transaction(tx, cred)

    def multi_start_roll(self, cred:Credentials, _hero_ids:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.multiStartRoll(_hero_ids)
        return self.send_transaction(tx, cred)

    def start_roll(self, cred:Credentials, _hero_id:uint256) -> TxReceipt:
        tx = self.contract.functions.startRoll(_hero_id)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xD5f5bE1037e457727e011ADE9Ca54d21c21a3F8A",
    "sd": "0x07520d5b2a7bf2DD0d48Bf08311Ac598F9ab4D4A"
}

ABI = """[
    {"name": "Approval", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "approved", "type": "address", "indexed": true, "internalType": "address"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "ApprovalForAll", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "operator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "approved", "type": "bool", "indexed": false, "internalType": "bool"}], "anonymous": false},
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "LandClaimed", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "landId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "region", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "LandMoved", "type": "event", "inputs": [{"name": "landId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "oldRegion", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "newRegion", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "Transfer", "type": "event", "inputs": [{"name": "from", "type": "address", "indexed": true, "internalType": "address"}, {"name": "to", "type": "address", "indexed": true, "internalType": "address"}, {"name": "tokenId", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "approve", "type": "function", "inputs": [{"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "balanceOf", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "claimLand", "type": "function", "inputs": [{"name": "_to", "type": "address", "internalType": "address"}, {"name": "_tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getAccountLands", "type": "function", "inputs": [{"name": "_account", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "landId", "type": "uint256", "internalType": "uint256"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "region", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint8", "internalType": "uint8"}, {"name": "steward", "type": "uint256", "internalType": "uint256"}, {"name": "score", "type": "uint64", "internalType": "uint64"}], "internalType": "struct LandCore.LandMeta[]"}], "stateMutability": "view"},
    {"name": "getAllLands", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "landId", "type": "uint256", "internalType": "uint256"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "region", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint8", "internalType": "uint8"}, {"name": "steward", "type": "uint256", "internalType": "uint256"}, {"name": "score", "type": "uint64", "internalType": "uint64"}], "internalType": "struct LandCore.LandMeta[]"}], "stateMutability": "view"},
    {"name": "getApproved", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "getLand", "type": "function", "inputs": [{"name": "_landId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "tuple", "components": [{"name": "landId", "type": "uint256", "internalType": "uint256"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "region", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint8", "internalType": "uint8"}, {"name": "steward", "type": "uint256", "internalType": "uint256"}, {"name": "score", "type": "uint64", "internalType": "uint64"}], "internalType": "struct LandCore.LandMeta"}], "stateMutability": "view"},
    {"name": "getLandsByRegion", "type": "function", "inputs": [{"name": "_region", "type": "uint32", "internalType": "uint32"}], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "landId", "type": "uint256", "internalType": "uint256"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "region", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint8", "internalType": "uint8"}, {"name": "steward", "type": "uint256", "internalType": "uint256"}, {"name": "score", "type": "uint64", "internalType": "uint64"}], "internalType": "struct LandCore.LandMeta[]"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "isApprovedForAll", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "operator", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "landIdToMeta", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "landId", "type": "uint256", "internalType": "uint256"}, {"name": "name", "type": "string", "internalType": "string"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "region", "type": "uint256", "internalType": "uint256"}, {"name": "level", "type": "uint8", "internalType": "uint8"}, {"name": "steward", "type": "uint256", "internalType": "uint256"}, {"name": "score", "type": "uint64", "internalType": "uint64"}], "stateMutability": "view"},
    {"name": "name", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "onERC721Received", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}, {"name": "", "type": "address", "internalType": "address"}, {"name": "", "type": "uint256", "internalType": "uint256"}, {"name": "", "type": "bytes", "internalType": "bytes"}], "outputs": [{"name": "", "type": "bytes4", "internalType": "bytes4"}], "stateMutability": "pure"},
    {"name": "ownerOf", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "pause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "regionToLandCount", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "regionToLands", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}, {"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "safeMint", "type": "function", "inputs": [{"name": "_owner", "type": "address", "internalType": "address"}, {"name": "_landId", "type": "uint256", "internalType": "uint256"}, {"name": "_name", "type": "string", "internalType": "string"}, {"name": "_region", "type": "uint32", "internalType": "uint32"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "safeTransferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "safeTransferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}, {"name": "_data", "type": "bytes", "internalType": "bytes"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setApprovalForAll", "type": "function", "inputs": [{"name": "operator", "type": "address", "internalType": "address"}, {"name": "approved", "type": "bool", "internalType": "bool"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "symbol", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "tokenByIndex", "type": "function", "inputs": [{"name": "index", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "tokenOfOwnerByIndex", "type": "function", "inputs": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "index", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "tokenURI", "type": "function", "inputs": [{"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "string", "internalType": "string"}], "stateMutability": "view"},
    {"name": "totalSupply", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "transferFrom", "type": "function", "inputs": [{"name": "from", "type": "address", "internalType": "address"}, {"name": "to", "type": "address", "internalType": "address"}, {"name": "tokenId", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "unpause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "updateLandRegion", "type": "function", "inputs": [{"name": "_landId", "type": "uint256", "internalType": "uint256"}, {"name": "_region", "type": "uint256", "internalType": "uint256"}, {"name": "_oldLandIndex", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class LandCore(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def approve(self, cred:Credentials, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.approve(to, token_id)
        return self.send_transaction(tx, cred)

    def balance_of(self, owner:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.balanceOf(owner).call(block_identifier=block_identifier)

    def claim_land(self, cred:Credentials, _to:address, _token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.claimLand(_to, _token_id)
        return self.send_transaction(tx, cred)

    def get_account_lands(self, _account:address, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getAccountLands(_account).call(block_identifier=block_identifier)

    def get_all_lands(self, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getAllLands().call(block_identifier=block_identifier)

    def get_approved(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.getApproved(token_id).call(block_identifier=block_identifier)

    def get_land(self, _land_id:uint256, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getLand(_land_id).call(block_identifier=block_identifier)

    def get_lands_by_region(self, _region:uint32, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getLandsByRegion(_region).call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.initialize()
        return self.send_transaction(tx, cred)

    def is_approved_for_all(self, owner:address, operator:address, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.isApprovedForAll(owner, operator).call(block_identifier=block_identifier)

    def land_id_to_meta(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, string, address, uint256, uint8, uint256, uint64]:
        return self.contract.functions.landIdToMeta(a).call(block_identifier=block_identifier)

    def name(self, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.name().call(block_identifier=block_identifier)

    def on_erc721_received(self, a:address, b:address, c:uint256, d:bytes, block_identifier:BlockIdentifier = 'latest') -> bytes4:
        return self.contract.functions.onERC721Received(a, b, c, d).call(block_identifier=block_identifier)

    def owner_of(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.ownerOf(token_id).call(block_identifier=block_identifier)

    def pause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.pause()
        return self.send_transaction(tx, cred)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def region_to_land_count(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.regionToLandCount(a).call(block_identifier=block_identifier)

    def region_to_lands(self, a:uint256, b:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.regionToLands(a, b).call(block_identifier=block_identifier)

    def safe_mint(self, cred:Credentials, _owner:address, _land_id:uint256, _name:string, _region:uint32) -> TxReceipt:
        tx = self.contract.functions.safeMint(_owner, _land_id, _name, _region)
        return self.send_transaction(tx, cred)

    def safe_transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.safeTransferFrom(_from, to, token_id)
        return self.send_transaction(tx, cred)

    def safe_transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256, _data:bytes) -> TxReceipt:
        tx = self.contract.functions.safeTransferFrom(_from, to, token_id, _data)
        return self.send_transaction(tx, cred)

    def set_approval_for_all(self, cred:Credentials, operator:address, approved:bool) -> TxReceipt:
        tx = self.contract.functions.setApprovalForAll(operator, approved)
        return self.send_transaction(tx, cred)

    def symbol(self, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.symbol().call(block_identifier=block_identifier)

    def token_by_index(self, index:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.tokenByIndex(index).call(block_identifier=block_identifier)

    def token_of_owner_by_index(self, owner:address, index:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.tokenOfOwnerByIndex(owner, index).call(block_identifier=block_identifier)

    def token_uri(self, token_id:uint256, block_identifier:BlockIdentifier = 'latest') -> string:
        return self.contract.functions.tokenURI(token_id).call(block_identifier=block_identifier)

    def total_supply(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalSupply().call(block_identifier=block_identifier)

    def transfer_from(self, cred:Credentials, _from:address, to:address, token_id:uint256) -> TxReceipt:
        tx = self.contract.functions.transferFrom(_from, to, token_id)
        return self.send_transaction(tx, cred)

    def unpause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.unpause()
        return self.send_transaction(tx, cred)

    def update_land_region(self, cred:Credentials, _land_id:uint256, _region:uint256, _old_land_index:uint256) -> TxReceipt:
        tx = self.contract.functions.updateLandRegion(_land_id, _region, _old_land_index)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x576C260513204392F0eC0bc865450872025CB1cA",
    "sd": "0x22e656419Be8A0abf0B53D0941FfDC3B70Fea36e"
}

ABI = """[
    {"name": "EggCracked", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "eggId", "type": "uint256", "indexed": true, "internalType": "uint256"}, {"name": "petId", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "EggIncubated", "type": "event", "inputs": [{"name": "owner", "type": "address", "indexed": true, "internalType": "address"}, {"name": "eggId", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "eggType", "type": "uint8", "indexed": true, "internalType": "uint8"}, {"name": "tier", "type": "uint8", "indexed": true, "internalType": "uint8"}], "anonymous": false},
    {"name": "FeeAddressAdded", "type": "event", "inputs": [{"name": "feeAddress", "type": "address", "indexed": true, "internalType": "address"}, {"name": "feePercent", "type": "uint256", "indexed": true, "internalType": "uint256"}], "anonymous": false},
    {"name": "FeeDeferred", "type": "event", "inputs": [{"name": "source", "type": "address", "indexed": true, "internalType": "address"}, {"name": "from", "type": "address", "indexed": true, "internalType": "address"}, {"name": "to", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token", "type": "address", "indexed": false, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "timestamp", "type": "uint64", "indexed": false, "internalType": "uint64"}], "anonymous": false},
    {"name": "FeeDisbursed", "type": "event", "inputs": [{"name": "source", "type": "address", "indexed": true, "internalType": "address"}, {"name": "from", "type": "address", "indexed": true, "internalType": "address"}, {"name": "to", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token", "type": "address", "indexed": false, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "timestamp", "type": "uint64", "indexed": false, "internalType": "uint64"}], "anonymous": false},
    {"name": "FeeLockedBurned", "type": "event", "inputs": [{"name": "source", "type": "address", "indexed": true, "internalType": "address"}, {"name": "from", "type": "address", "indexed": true, "internalType": "address"}, {"name": "to", "type": "address", "indexed": true, "internalType": "address"}, {"name": "token", "type": "address", "indexed": false, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "timestamp", "type": "uint64", "indexed": false, "internalType": "uint64"}], "anonymous": false},
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "RoleAdminChanged", "type": "event", "inputs": [{"name": "role", "type": "bytes32", "indexed": true, "internalType": "bytes32"}, {"name": "previousAdminRole", "type": "bytes32", "indexed": true, "internalType": "bytes32"}, {"name": "newAdminRole", "type": "bytes32", "indexed": true, "internalType": "bytes32"}], "anonymous": false},
    {"name": "RoleGranted", "type": "event", "inputs": [{"name": "role", "type": "bytes32", "indexed": true, "internalType": "bytes32"}, {"name": "account", "type": "address", "indexed": true, "internalType": "address"}, {"name": "sender", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "RoleRevoked", "type": "event", "inputs": [{"name": "role", "type": "bytes32", "indexed": true, "internalType": "bytes32"}, {"name": "account", "type": "address", "indexed": true, "internalType": "address"}, {"name": "sender", "type": "address", "indexed": true, "internalType": "address"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "DEFAULT_ADMIN_ROLE", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "HATCHING_MODERATOR_ROLE", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "MODERATOR_ROLE", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "crack", "type": "function", "inputs": [{"name": "_eggId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "nonpayable"},
    {"name": "eggTypeCosts", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "eggAddress", "type": "address", "internalType": "address"}, {"name": "itemAddress1", "type": "address", "internalType": "address"}, {"name": "itemAmount1", "type": "uint16", "internalType": "uint16"}, {"name": "itemAddress2", "type": "address", "internalType": "address"}, {"name": "itemAmount2", "type": "uint16", "internalType": "uint16"}], "stateMutability": "view"},
    {"name": "feeAddresses", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "feePercents", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getEgg", "type": "function", "inputs": [{"name": "_eggId", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "tuple", "components": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "petId", "type": "uint256", "internalType": "uint256"}, {"name": "owner", "type": "address", "internalType": "address"}, {"name": "eggType", "type": "uint8", "internalType": "uint8"}, {"name": "seedblock", "type": "uint256", "internalType": "uint256"}, {"name": "finishTime", "type": "uint256", "internalType": "uint256"}, {"name": "tier", "type": "uint8", "internalType": "uint8"}], "internalType": "struct UnhatchedEgg"}], "stateMutability": "view"},
    {"name": "getRoleAdmin", "type": "function", "inputs": [{"name": "role", "type": "bytes32", "internalType": "bytes32"}], "outputs": [{"name": "", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "getUserEggs", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "grantRole", "type": "function", "inputs": [{"name": "role", "type": "bytes32", "internalType": "bytes32"}, {"name": "account", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "hasRole", "type": "function", "inputs": [{"name": "role", "type": "bytes32", "internalType": "bytes32"}, {"name": "account", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "incubateEgg", "type": "function", "inputs": [{"name": "_eggType", "type": "uint8", "internalType": "uint8"}, {"name": "_tier", "type": "uint8", "internalType": "uint8"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "incubateEggWithLocked", "type": "function", "inputs": [{"name": "_eggType", "type": "uint8", "internalType": "uint8"}, {"name": "_tier", "type": "uint8", "internalType": "uint8"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_petCoreAddress", "type": "address", "internalType": "address"}, {"name": "_powerTokenAddress", "type": "address", "internalType": "address"}, {"name": "_goldAddress", "type": "address", "internalType": "address"}, {"name": "_gaiaTearsAddress", "type": "address", "internalType": "address"}, {"name": "_randomGeneratorAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "originId", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint8", "internalType": "uint8"}], "stateMutability": "view"},
    {"name": "pause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "powerToken", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "contract IPowerToken"}], "stateMutability": "view"},
    {"name": "priceTiers", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "powerTokenCost", "type": "uint8", "internalType": "uint8"}, {"name": "goldCost", "type": "uint16", "internalType": "uint16"}, {"name": "tearCost", "type": "uint8", "internalType": "uint8"}, {"name": "incubationTime", "type": "uint32", "internalType": "uint32"}, {"name": "shinyChance", "type": "uint16", "internalType": "uint16"}], "stateMutability": "view"},
    {"name": "renounceRole", "type": "function", "inputs": [{"name": "role", "type": "bytes32", "internalType": "bytes32"}, {"name": "account", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "revokeRole", "type": "function", "inputs": [{"name": "role", "type": "bytes32", "internalType": "bytes32"}, {"name": "account", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "season", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint8", "internalType": "uint8"}], "stateMutability": "view"},
    {"name": "setAppearanceChoices", "type": "function", "inputs": [{"name": "_eggType", "type": "uint8", "internalType": "uint8"}, {"name": "_rarity", "type": "uint8", "internalType": "uint8"}, {"name": "_isSpecial", "type": "uint8", "internalType": "uint8"}, {"name": "_startIndex", "type": "uint256", "internalType": "uint256"}, {"name": "_endIndex", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setEggTypeCosts", "type": "function", "inputs": [{"name": "_eggType", "type": "uint8", "internalType": "uint8"}, {"name": "_eggTypeCost", "type": "tuple", "components": [{"name": "eggAddress", "type": "address", "internalType": "address"}, {"name": "itemAddress1", "type": "address", "internalType": "address"}, {"name": "itemAmount1", "type": "uint16", "internalType": "uint16"}, {"name": "itemAddress2", "type": "address", "internalType": "address"}, {"name": "itemAmount2", "type": "uint16", "internalType": "uint16"}], "internalType": "struct EggTypeCost"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setFees", "type": "function", "inputs": [{"name": "_feeAddresses", "type": "address[]", "internalType": "address[]"}, {"name": "_feePercents", "type": "uint256[]", "internalType": "uint256[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setOriginId", "type": "function", "inputs": [{"name": "_originId", "type": "uint8", "internalType": "uint8"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setPetCore", "type": "function", "inputs": [{"name": "_petCoreAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setPriceTiers", "type": "function", "inputs": [{"name": "_priceTierIndex", "type": "uint8", "internalType": "uint8"}, {"name": "_priceTier", "type": "tuple", "components": [{"name": "powerTokenCost", "type": "uint8", "internalType": "uint8"}, {"name": "goldCost", "type": "uint16", "internalType": "uint16"}, {"name": "tearCost", "type": "uint8", "internalType": "uint8"}, {"name": "incubationTime", "type": "uint32", "internalType": "uint32"}, {"name": "shinyChance", "type": "uint16", "internalType": "uint16"}], "internalType": "struct PriceTier"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setTokenUnlocker", "type": "function", "inputs": [{"name": "_tokenUnlockerAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "supportsInterface", "type": "function", "inputs": [{"name": "interfaceId", "type": "bytes4", "internalType": "bytes4"}], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "totalEggs", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "unpause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class PetHatching(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def crack(self, cred:Credentials, _egg_id:uint256) -> TxReceipt:
        tx = self.contract.functions.crack(_egg_id)
        return self.send_transaction(tx, cred)

    def egg_type_costs(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[address, address, uint16, address, uint16]:
        return self.contract.functions.eggTypeCosts(a).call(block_identifier=block_identifier)

    def fee_addresses(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.feeAddresses(a).call(block_identifier=block_identifier)

    def fee_percents(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.feePercents(a).call(block_identifier=block_identifier)

    def get_egg(self, _egg_id:uint256, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getEgg(_egg_id).call(block_identifier=block_identifier)

    def get_user_eggs(self, _address:address, block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getUserEggs(_address).call(block_identifier=block_identifier)

    def incubate_egg(self, cred:Credentials, _egg_type:uint8, _tier:uint8) -> TxReceipt:
        tx = self.contract.functions.incubateEgg(_egg_type, _tier)
        return self.send_transaction(tx, cred)

    def incubate_egg_with_locked(self, cred:Credentials, _egg_type:uint8, _tier:uint8) -> TxReceipt:
        tx = self.contract.functions.incubateEggWithLocked(_egg_type, _tier)
        return self.send_transaction(tx, cred)

    def initialize(self, cred:Credentials, _pet_core_address:address, _power_token_address:address, _gold_address:address, _gaia_tears_address:address, _random_generator_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_pet_core_address, _power_token_address, _gold_address, _gaia_tears_address, _random_generator_address)
        return self.send_transaction(tx, cred)

    def origin_id(self, block_identifier:BlockIdentifier = 'latest') -> uint8:
        return self.contract.functions.originId().call(block_identifier=block_identifier)

    def pause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.pause()
        return self.send_transaction(tx, cred)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def power_token(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.powerToken().call(block_identifier=block_identifier)

    def price_tiers(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint8, uint16, uint8, uint32, uint16]:
        return self.contract.functions.priceTiers(a).call(block_identifier=block_identifier)

    def season(self, block_identifier:BlockIdentifier = 'latest') -> uint8:
        return self.contract.functions.season().call(block_identifier=block_identifier)

    def set_appearance_choices(self, cred:Credentials, _egg_type:uint8, _rarity:uint8, _is_special:uint8, _start_index:uint256, _end_index:uint256) -> TxReceipt:
        tx = self.contract.functions.setAppearanceChoices(_egg_type, _rarity, _is_special, _start_index, _end_index)
        return self.send_transaction(tx, cred)

    def set_egg_type_costs(self, cred:Credentials, _egg_type:uint8, _egg_type_cost:tuple) -> TxReceipt:
        tx = self.contract.functions.setEggTypeCosts(_egg_type, _egg_type_cost)
        return self.send_transaction(tx, cred)

    def set_fees(self, cred:Credentials, _fee_addresses:Sequence[address], _fee_percents:Sequence[uint256]) -> TxReceipt:
        tx = self.contract.functions.setFees(_fee_addresses, _fee_percents)
        return self.send_transaction(tx, cred)

    def set_origin_id(self, cred:Credentials, _origin_id:uint8) -> TxReceipt:
        tx = self.contract.functions.setOriginId(_origin_id)
        return self.send_transaction(tx, cred)

    def set_pet_core(self, cred:Credentials, _pet_core_address:address) -> TxReceipt:
        tx = self.contract.functions.setPetCore(_pet_core_address)
        return self.send_transaction(tx, cred)

    def set_price_tiers(self, cred:Credentials, _price_tier_index:uint8, _price_tier:tuple) -> TxReceipt:
        tx = self.contract.functions.setPriceTiers(_price_tier_index, _price_tier)
        return self.send_transaction(tx, cred)

    def set_token_unlocker(self, cred:Credentials, _token_unlocker_address:address) -> TxReceipt:
        tx = self.contract.functions.setTokenUnlocker(_token_unlocker_address)
        return self.send_transaction(tx, cred)

    def supports_interface(self, interface_id:bytes4, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.supportsInterface(interface_id).call(block_identifier=block_identifier)

    def total_eggs(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalEggs().call(block_identifier=block_identifier)

    def unpause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.unpause()
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xcA11bde05977b3631167028862bE2a173976CA11",
    "sd": "0xcA11bde05977b3631167028862bE2a173976CA11"
}

ABI = """[
    {"name": "aggregate", "type": "function", "inputs": [{"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call[]"}], "outputs": [{"name": "blockNumber", "type": "uint256", "internalType": "uint256"}, {"name": "returnData", "type": "bytes[]", "internalType": "bytes[]"}], "stateMutability": "view"},
    {"name": "aggregate3", "type": "function", "inputs": [{"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "allowFailure", "type": "bool", "internalType": "bool"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call3[]"}], "outputs": [{"name": "returnData", "type": "tuple[]", "components": [{"name": "success", "type": "bool", "internalType": "bool"}, {"name": "returnData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Result[]"}], "stateMutability": "view"},
    {"name": "aggregate3Value", "type": "function", "inputs": [{"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "allowFailure", "type": "bool", "internalType": "bool"}, {"name": "value", "type": "uint256", "internalType": "uint256"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call3Value[]"}], "outputs": [{"name": "returnData", "type": "tuple[]", "components": [{"name": "success", "type": "bool", "internalType": "bool"}, {"name": "returnData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Result[]"}], "stateMutability": "payable"},
    {"name": "blockAndAggregate", "type": "function", "inputs": [{"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call[]"}], "outputs": [{"name": "blockNumber", "type": "uint256", "internalType": "uint256"}, {"name": "blockHash", "type": "bytes32", "internalType": "bytes32"}, {"name": "returnData", "type": "tuple[]", "components": [{"name": "success", "type": "bool", "internalType": "bool"}, {"name": "returnData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Result[]"}], "stateMutability": "view"},
    {"name": "getBasefee", "type": "function", "inputs": [], "outputs": [{"name": "basefee", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getBlockHash", "type": "function", "inputs": [{"name": "blockNumber", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "blockHash", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "getBlockNumber", "type": "function", "inputs": [], "outputs": [{"name": "blockNumber", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getChainId", "type": "function", "inputs": [], "outputs": [{"name": "chainid", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getCurrentBlockCoinbase", "type": "function", "inputs": [], "outputs": [{"name": "coinbase", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"name": "getCurrentBlockDifficulty", "type": "function", "inputs": [], "outputs": [{"name": "difficulty", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getCurrentBlockGasLimit", "type": "function", "inputs": [], "outputs": [{"name": "gaslimit", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getCurrentBlockTimestamp", "type": "function", "inputs": [], "outputs": [{"name": "timestamp", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getEthBalance", "type": "function", "inputs": [{"name": "addr", "type": "address", "internalType": "address"}], "outputs": [{"name": "balance", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getLastBlockHash", "type": "function", "inputs": [], "outputs": [{"name": "blockHash", "type": "bytes32", "internalType": "bytes32"}], "stateMutability": "view"},
    {"name": "tryAggregate", "type": "function", "inputs": [{"name": "requireSuccess", "type": "bool", "internalType": "bool"}, {"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call[]"}], "outputs": [{"name": "returnData", "type": "tuple[]", "components": [{"name": "success", "type": "bool", "internalType": "bool"}, {"name": "returnData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Result[]"}], "stateMutability": "view"},
    {"name": "tryBlockAndAggregate", "type": "function", "inputs": [{"name": "requireSuccess", "type": "bool", "internalType": "bool"}, {"name": "calls", "type": "tuple[]", "components": [{"name": "target", "type": "address", "internalType": "address"}, {"name": "callData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Call[]"}], "outputs": [{"name": "blockNumber", "type": "uint256", "internalType": "uint256"}, {"name": "blockHash", "type": "bytes32", "internalType": "bytes32"}, {"name": "returnData", "type": "tuple[]", "components": [{"name": "success", "type": "bool", "internalType": "bool"}, {"name": "returnData", "type": "bytes", "internalType": "bytes"}], "internalType": "struct Multicall3.Result[]"}], "stateMutability": "nonpayable"}
]
"""     

class Multicall3(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def aggregate(self, calls:Sequence[tuple], block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, List[bytes]]:
        return self.contract.functions.aggregate(calls).call(block_identifier=block_identifier)

    def aggregate3(self, calls:Sequence[tuple], block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.aggregate3(calls).call(block_identifier=block_identifier)

    def aggregate3_value(self, cred:Credentials, calls:Sequence[tuple]) -> TxReceipt:
        tx = self.contract.functions.aggregate3Value(calls)
        return self.send_transaction(tx, cred)

    def block_and_aggregate(self, calls:Sequence[tuple], block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, bytes32, List[tuple]]:
        return self.contract.functions.blockAndAggregate(calls).call(block_identifier=block_identifier)

    def get_basefee(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getBasefee().call(block_identifier=block_identifier)

    def get_block_hash(self, block_number:uint256, block_identifier:BlockIdentifier = 'latest') -> bytes32:
        return self.contract.functions.getBlockHash(block_number).call(block_identifier=block_identifier)

    def get_block_number(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getBlockNumber().call(block_identifier=block_identifier)

    def get_chain_id(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getChainId().call(block_identifier=block_identifier)

    def get_current_block_coinbase(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.getCurrentBlockCoinbase().call(block_identifier=block_identifier)

    def get_current_block_difficulty(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getCurrentBlockDifficulty().call(block_identifier=block_identifier)

    def get_current_block_gas_limit(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getCurrentBlockGasLimit().call(block_identifier=block_identifier)

    def get_current_block_timestamp(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getCurrentBlockTimestamp().call(block_identifier=block_identifier)

    def get_eth_balance(self, addr:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.getEthBalance(addr).call(block_identifier=block_identifier)

    def get_last_block_hash(self, block_identifier:BlockIdentifier = 'latest') -> bytes32:
        return self.contract.functions.getLastBlockHash().call(block_identifier=block_identifier)

    def try_aggregate(self, require_success:bool, calls:Sequence[tuple], block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.tryAggregate(require_success, calls).call(block_identifier=block_identifier)

    def try_block_and_aggregate(self, cred:Credentials, require_success:bool, calls:Sequence[tuple]) -> TxReceipt:
        tx = self.contract.functions.tryBlockAndAggregate(require_success, calls)
        return self.send_transaction(tx, cred)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x87CBa8F998F902f2fff990efFa1E261F35932e57",
    "sd": "0x696BECc6ddD5589a9a3Bb93fAc3A53D26b7cb819"
}

ABI = """[
    {"name": "Initialized", "type": "event", "inputs": [{"name": "version", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "Paused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "PotionAdded", "type": "event", "inputs": [{"name": "potionAddress", "type": "address", "indexed": true, "internalType": "address"}, {"name": "requiredResources", "type": "address[]", "indexed": false, "internalType": "address[]"}, {"name": "requiredQuantities", "type": "uint32[]", "indexed": false, "internalType": "uint32[]"}], "anonymous": false},
    {"name": "PotionCreated", "type": "event", "inputs": [{"name": "player", "type": "address", "indexed": true, "internalType": "address"}, {"name": "potionAddress", "type": "address", "indexed": false, "internalType": "address"}, {"name": "quantity", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "requiredResources", "type": "address[]", "indexed": false, "internalType": "address[]"}, {"name": "requiredQuantities", "type": "uint32[]", "indexed": false, "internalType": "uint32[]"}], "anonymous": false},
    {"name": "PotionUpdated", "type": "event", "inputs": [{"name": "potionAddress", "type": "address", "indexed": true, "internalType": "address"}, {"name": "requiredResources", "type": "address[]", "indexed": false, "internalType": "address[]"}, {"name": "requiredQuantities", "type": "uint32[]", "indexed": false, "internalType": "uint32[]"}, {"name": "status", "type": "uint8", "indexed": false, "internalType": "uint8"}], "anonymous": false},
    {"name": "Unpaused", "type": "event", "inputs": [{"name": "account", "type": "address", "indexed": false, "internalType": "address"}], "anonymous": false},
    {"name": "addPotion", "type": "function", "inputs": [{"name": "_potionAddress", "type": "address", "internalType": "address"}, {"name": "_requiredResources", "type": "address[]", "internalType": "address[]"}, {"name": "_requiredQuantities", "type": "uint32[]", "internalType": "uint32[]"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "addressToPotionId", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "createPotion", "type": "function", "inputs": [{"name": "_potionAddress", "type": "address", "internalType": "address"}, {"name": "_quantity", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "getPotion", "type": "function", "inputs": [{"name": "_potionAddress", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "tuple", "components": [{"name": "potionAddress", "type": "address", "internalType": "address"}, {"name": "requiredResources", "type": "address[]", "internalType": "address[]"}, {"name": "requiredQuantities", "type": "uint32[]", "internalType": "uint32[]"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "internalType": "struct Alchemist.Potion"}], "stateMutability": "view"},
    {"name": "getPotions", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "potionAddress", "type": "address", "internalType": "address"}, {"name": "requiredResources", "type": "address[]", "internalType": "address[]"}, {"name": "requiredQuantities", "type": "uint32[]", "internalType": "uint32[]"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "internalType": "struct Alchemist.Potion[]"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_dfkGoldAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "paused", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "potions", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "potionAddress", "type": "address", "internalType": "address"}, {"name": "status", "type": "uint8", "internalType": "uint8"}], "stateMutability": "view"},
    {"name": "removePotion", "type": "function", "inputs": [{"name": "_potionAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "togglePause", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "updatePotion", "type": "function", "inputs": [{"name": "_potionAddress", "type": "address", "internalType": "address"}, {"name": "_requiredResources", "type": "address[]", "internalType": "address[]"}, {"name": "_requiredQuantities", "type": "uint32[]", "internalType": "uint32[]"}, {"name": "_status", "type": "uint8", "internalType": "uint8"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class Alchemist(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def add_potion(self, cred:Credentials, _potion_address:address, _required_resources:Sequence[address], _required_quantities:Sequence[uint32]) -> TxReceipt:
        tx = self.contract.functions.addPotion(_potion_address, _required_resources, _required_quantities)
        return self.send_transaction(tx, cred)

    def address_to_potion_id(self, a:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.addressToPotionId(a).call(block_identifier=block_identifier)

    def create_potion(self, cred:Credentials, _potion_address:address, _quantity:uint256) -> TxReceipt:
        tx = self.contract.functions.createPotion(_potion_address, _quantity)
        return self.send_transaction(tx, cred)

    def get_potion(self, _potion_address:address, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getPotion(_potion_address).call(block_identifier=block_identifier)

    def get_potions(self, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getPotions().call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials, _dfk_gold_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_dfk_gold_address)
        return self.send_transaction(tx, cred)

    def paused(self, block_identifier:BlockIdentifier = 'latest') -> bool:
        return self.contract.functions.paused().call(block_identifier=block_identifier)

    def potions(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[address, uint8]:
        return self.contract.functions.potions(a).call(block_identifier=block_identifier)

    def remove_potion(self, cred:Credentials, _potion_address:address) -> TxReceipt:
        tx = self.contract.functions.removePotion(_potion_address)
        return self.send_transaction(tx, cred)

    def toggle_pause(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.togglePause()
        return self.send_transaction(tx, cred)

    def update_potion(self, cred:Credentials, _potion_address:address, _required_resources:Sequence[address], _required_quantities:Sequence[uint32], _status:uint8) -> TxReceipt:
        tx = self.contract.functions.updatePotion(_potion_address, _required_resources, _required_quantities, _status)
        return self.send_transaction(tx, cred)


#! /usr/bin/env python

from .solidity_types import address as HexAddress
from typing import Dict, List, Optional, Sequence, Tuple, Union, cast

class Credentials:
    def __init__(self, address:HexAddress|str, private_key:str ='', nickname:str=''):
        self.address:HexAddress = cast(HexAddress, address)
        self.private_key = private_key
        self.nickname = nickname

    def short_address(self):
        return f'{self.address[0:6]}…{self.address[-4:]}'

    def __repr__(self):
        return f'{self.nickname:<10}:  {self.address}'

    def matches_abbreviation(self, query:str) -> bool:
        # An 'abbreviation' is a string that has the same beginning 
        # as self.nickname, case-insensitive, e.g. 'sli' for 'Slipfoot'
        return self.nickname.lower().startswith(query.lower())

    @staticmethod
    def cred_for_nickname(creds:Sequence['Credentials'], nickname:str, accept_abbreviation=True) -> 'Credentials':
        if accept_abbreviation:
            gen = (c for c in creds if c.matches_abbreviation(nickname))
        else:
            gen = (c for c in creds if c.nickname == nickname)
        try:
            return next(gen)
        except StopIteration:
            raise ValueError(f'Nickname {nickname} not found in {creds}')

    @staticmethod
    def cred_for_address(creds:Sequence['Credentials'], address:HexAddress) -> 'Credentials':
        gen = (c for c in creds if c.address == address)
        try:
            return next(gen)
        except StopIteration:
            raise ValueError(f'Address {address} not found in {creds}')   


    @staticmethod 
    def creds_for_nicknames(creds:Sequence['Credentials'], nicknames:Sequence[str], accept_abbreviation=True) -> List['Credentials']:
        creds_out = []
        for n in nicknames:
            try:
                c = Credentials.cred_for_nickname(creds, n, accept_abbreviation)
                creds_out.append(c)
            except ValueError:
                pass
        return creds_out




from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0xED6dC9FD092190C08e4afF8611496774Ded19D54",
    "sd": "0x0000000000000000000000000000000000000000"
}

ABI = """[
    {"name": "AddValidator", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": false, "internalType": "address"}, {"name": "nodeID", "type": "string", "indexed": false, "internalType": "string"}], "anonymous": false},
    {"name": "Disbursement", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Process", "type": "event", "inputs": [{"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "RemoveValidator", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": false, "internalType": "address"}, {"name": "nodeID", "type": "string", "indexed": false, "internalType": "string"}], "anonymous": false},
    {"name": "Slash", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Stake", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "endTime", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "Unstake", "type": "event", "inputs": [{"name": "validator", "type": "address", "indexed": true, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "INITIAL_MAX_STAKE_AMOUNT", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "INITIAL_MAX_STAKE_DURATION", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "INITIAL_MIN_STAKE_AMOUNT", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "INITIAL_MIN_STAKE_DURATION", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "addValidator", "type": "function", "inputs": [{"name": "_address", "type": "address", "internalType": "address"}, {"name": "_nodeID", "type": "string", "internalType": "string"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "addressToValidator", "type": "function", "inputs": [{"name": "", "type": "address", "internalType": "address"}], "outputs": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "nodeID", "type": "string", "internalType": "string"}, {"name": "rewardAddress", "type": "address", "internalType": "address"}, {"name": "stakeAmount", "type": "uint256", "internalType": "uint256"}, {"name": "stakeStartAt", "type": "uint256", "internalType": "uint256"}, {"name": "stakeEndAt", "type": "uint256", "internalType": "uint256"}, {"name": "balance", "type": "uint256", "internalType": "uint256"}, {"name": "lifetimeBalance", "type": "uint256", "internalType": "uint256"}, {"name": "exists", "type": "bool", "internalType": "bool"}], "stateMutability": "view"},
    {"name": "allocatedAmount", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "burnPercentage", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "claimBalance", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "fundAddress", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "address payable"}], "stateMutability": "view"},
    {"name": "fundPercentage", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "getValidator", "type": "function", "inputs": [{"name": "_validator", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "tuple", "components": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "nodeID", "type": "string", "internalType": "string"}, {"name": "rewardAddress", "type": "address", "internalType": "address"}, {"name": "stakeAmount", "type": "uint256", "internalType": "uint256"}, {"name": "stakeStartAt", "type": "uint256", "internalType": "uint256"}, {"name": "stakeEndAt", "type": "uint256", "internalType": "uint256"}, {"name": "balance", "type": "uint256", "internalType": "uint256"}, {"name": "lifetimeBalance", "type": "uint256", "internalType": "uint256"}, {"name": "exists", "type": "bool", "internalType": "bool"}], "internalType": "struct ValidatorFund.Validator"}], "stateMutability": "view"},
    {"name": "getValidators", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "tuple[]", "components": [{"name": "owner", "type": "address", "internalType": "address"}, {"name": "nodeID", "type": "string", "internalType": "string"}, {"name": "rewardAddress", "type": "address", "internalType": "address"}, {"name": "stakeAmount", "type": "uint256", "internalType": "uint256"}, {"name": "stakeStartAt", "type": "uint256", "internalType": "uint256"}, {"name": "stakeEndAt", "type": "uint256", "internalType": "uint256"}, {"name": "balance", "type": "uint256", "internalType": "uint256"}, {"name": "lifetimeBalance", "type": "uint256", "internalType": "uint256"}, {"name": "exists", "type": "bool", "internalType": "bool"}], "internalType": "struct ValidatorFund.Validator[]"}], "stateMutability": "view"},
    {"name": "govToken", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "address", "internalType": "contract IERC20"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_burnPercentage", "type": "uint256", "internalType": "uint256"}, {"name": "_jewelFundPercentage", "type": "uint256", "internalType": "uint256"}, {"name": "_validatorFundPercentage", "type": "uint256", "internalType": "uint256"}, {"name": "_govTokenAddress", "type": "address", "internalType": "address"}, {"name": "_fundAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "lastDistribution", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "maxStakeAmount", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "maxStakeDuration", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "minStakeAmount", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "minStakeDuration", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "processFunds", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "removeValidator", "type": "function", "inputs": [{"name": "_validator", "type": "address", "internalType": "address"}, {"name": "_index", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "rewardPercentage", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "setPercentages", "type": "function", "inputs": [{"name": "_burnPercentage", "type": "uint256", "internalType": "uint256"}, {"name": "_fundPercentage", "type": "uint256", "internalType": "uint256"}, {"name": "_rewardPercentage", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setStakeAmounts", "type": "function", "inputs": [{"name": "_minStakeAmount", "type": "uint256", "internalType": "uint256"}, {"name": "_maxStakeAmount", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "setStakeDurations", "type": "function", "inputs": [{"name": "_minStakeDuration", "type": "uint256", "internalType": "uint256"}, {"name": "_maxStakeDuration", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "slash", "type": "function", "inputs": [{"name": "_validator", "type": "address", "internalType": "address"}, {"name": "_wei", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "stake", "type": "function", "inputs": [{"name": "_amount", "type": "uint256", "internalType": "uint256"}, {"name": "_duration", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "totalBurn", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalFund", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalReward", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalStake", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "unallocated", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "unstake", "type": "function", "inputs": [], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "updateRewardAddress", "type": "function", "inputs": [{"name": "_rewardAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "validators", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "", "type": "address", "internalType": "address"}], "stateMutability": "view"},
    {"type": "receive", "stateMutability": "payable"}
]
"""     

class ValidatorFund(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def initial_max_stake_amount(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.INITIAL_MAX_STAKE_AMOUNT().call(block_identifier=block_identifier)

    def initial_max_stake_duration(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.INITIAL_MAX_STAKE_DURATION().call(block_identifier=block_identifier)

    def initial_min_stake_amount(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.INITIAL_MIN_STAKE_AMOUNT().call(block_identifier=block_identifier)

    def initial_min_stake_duration(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.INITIAL_MIN_STAKE_DURATION().call(block_identifier=block_identifier)

    def add_validator(self, cred:Credentials, _address:address, _node_id:string) -> TxReceipt:
        tx = self.contract.functions.addValidator(_address, _node_id)
        return self.send_transaction(tx, cred)

    def address_to_validator(self, a:address, block_identifier:BlockIdentifier = 'latest') -> Tuple[address, string, address, uint256, uint256, uint256, uint256, uint256, bool]:
        return self.contract.functions.addressToValidator(a).call(block_identifier=block_identifier)

    def allocated_amount(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.allocatedAmount().call(block_identifier=block_identifier)

    def burn_percentage(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.burnPercentage().call(block_identifier=block_identifier)

    def claim_balance(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.claimBalance()
        return self.send_transaction(tx, cred)

    def fund_address(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.fundAddress().call(block_identifier=block_identifier)

    def fund_percentage(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.fundPercentage().call(block_identifier=block_identifier)

    def get_validator(self, _validator:address, block_identifier:BlockIdentifier = 'latest') -> tuple:
        return self.contract.functions.getValidator(_validator).call(block_identifier=block_identifier)

    def get_validators(self, block_identifier:BlockIdentifier = 'latest') -> List[tuple]:
        return self.contract.functions.getValidators().call(block_identifier=block_identifier)

    def gov_token(self, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.govToken().call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials, _burn_percentage:uint256, _jewel_fund_percentage:uint256, _validator_fund_percentage:uint256, _gov_token_address:address, _fund_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_burn_percentage, _jewel_fund_percentage, _validator_fund_percentage, _gov_token_address, _fund_address)
        return self.send_transaction(tx, cred)

    def last_distribution(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.lastDistribution().call(block_identifier=block_identifier)

    def max_stake_amount(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.maxStakeAmount().call(block_identifier=block_identifier)

    def max_stake_duration(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.maxStakeDuration().call(block_identifier=block_identifier)

    def min_stake_amount(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.minStakeAmount().call(block_identifier=block_identifier)

    def min_stake_duration(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.minStakeDuration().call(block_identifier=block_identifier)

    def process_funds(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.processFunds()
        return self.send_transaction(tx, cred)

    def remove_validator(self, cred:Credentials, _validator:address, _index:uint256) -> TxReceipt:
        tx = self.contract.functions.removeValidator(_validator, _index)
        return self.send_transaction(tx, cred)

    def reward_percentage(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.rewardPercentage().call(block_identifier=block_identifier)

    def set_percentages(self, cred:Credentials, _burn_percentage:uint256, _fund_percentage:uint256, _reward_percentage:uint256) -> TxReceipt:
        tx = self.contract.functions.setPercentages(_burn_percentage, _fund_percentage, _reward_percentage)
        return self.send_transaction(tx, cred)

    def set_stake_amounts(self, cred:Credentials, _min_stake_amount:uint256, _max_stake_amount:uint256) -> TxReceipt:
        tx = self.contract.functions.setStakeAmounts(_min_stake_amount, _max_stake_amount)
        return self.send_transaction(tx, cred)

    def set_stake_durations(self, cred:Credentials, _min_stake_duration:uint256, _max_stake_duration:uint256) -> TxReceipt:
        tx = self.contract.functions.setStakeDurations(_min_stake_duration, _max_stake_duration)
        return self.send_transaction(tx, cred)

    def slash(self, cred:Credentials, _validator:address, _wei:uint256) -> TxReceipt:
        tx = self.contract.functions.slash(_validator, _wei)
        return self.send_transaction(tx, cred)

    def stake(self, cred:Credentials, _amount:uint256, _duration:uint256) -> TxReceipt:
        tx = self.contract.functions.stake(_amount, _duration)
        return self.send_transaction(tx, cred)

    def total_burn(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalBurn().call(block_identifier=block_identifier)

    def total_fund(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalFund().call(block_identifier=block_identifier)

    def total_reward(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalReward().call(block_identifier=block_identifier)

    def total_stake(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalStake().call(block_identifier=block_identifier)

    def unallocated(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.unallocated().call(block_identifier=block_identifier)

    def unstake(self, cred:Credentials) -> TxReceipt:
        tx = self.contract.functions.unstake()
        return self.send_transaction(tx, cred)

    def update_reward_address(self, cred:Credentials, _reward_address:address) -> TxReceipt:
        tx = self.contract.functions.updateRewardAddress(_reward_address)
        return self.send_transaction(tx, cred)

    def validators(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> address:
        return self.contract.functions.validators(a).call(block_identifier=block_identifier)



from ..abi_contract_wrapper import ABIContractWrapper
from ..solidity_types import *
from ..credentials import Credentials

CONTRACT_ADDRESS =     {
    "cv": "0x123165B3a30fdA3655B30cfC10135C1CA3C21bFC",
    "sd": "0x0000000000000000000000000000000000000000"
}

ABI = """[
    {"name": "DisbursementAdded", "type": "event", "inputs": [{"name": "id", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "recipient", "type": "address", "indexed": false, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "startTime", "type": "uint64", "indexed": false, "internalType": "uint64"}, {"name": "duration", "type": "uint64", "indexed": false, "internalType": "uint64"}], "anonymous": false},
    {"name": "DisbursementClaim", "type": "event", "inputs": [{"name": "id", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "recipient", "type": "address", "indexed": false, "internalType": "address"}, {"name": "amount", "type": "uint256", "indexed": false, "internalType": "uint256"}], "anonymous": false},
    {"name": "DisbursementUpdated", "type": "event", "inputs": [{"name": "id", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "reduction", "type": "uint256", "indexed": false, "internalType": "uint256"}, {"name": "startTime", "type": "uint64", "indexed": false, "internalType": "uint64"}, {"name": "duration", "type": "uint64", "indexed": false, "internalType": "uint64"}], "anonymous": false},
    {"name": "addDisbursement", "type": "function", "inputs": [{"name": "_recipient", "type": "address", "internalType": "address"}, {"name": "_amount", "type": "uint256", "internalType": "uint256"}, {"name": "_startTime", "type": "uint64", "internalType": "uint64"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}, {"name": "_note", "type": "string", "internalType": "string"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "addDisbursements", "type": "function", "inputs": [{"name": "_recipients", "type": "address[]", "internalType": "address[]"}, {"name": "_amounts", "type": "uint256[]", "internalType": "uint256[]"}, {"name": "_startTime", "type": "uint64", "internalType": "uint64"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}, {"name": "_note", "type": "string", "internalType": "string"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "claim", "type": "function", "inputs": [{"name": "_amount", "type": "uint256", "internalType": "uint256"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "disbursements", "type": "function", "inputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "outputs": [{"name": "id", "type": "uint256", "internalType": "uint256"}, {"name": "recipient", "type": "address", "internalType": "address"}, {"name": "note", "type": "string", "internalType": "string"}, {"name": "total", "type": "uint256", "internalType": "uint256"}, {"name": "claimed", "type": "uint256", "internalType": "uint256"}, {"name": "balance", "type": "uint256", "internalType": "uint256"}, {"name": "createdTime", "type": "uint64", "internalType": "uint64"}, {"name": "startTime", "type": "uint64", "internalType": "uint64"}, {"name": "duration", "type": "uint64", "internalType": "uint64"}], "stateMutability": "view"},
    {"name": "getDisbursementIds", "type": "function", "inputs": [{"name": "_recipient", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256[]", "internalType": "uint256[]"}], "stateMutability": "view"},
    {"name": "initialize", "type": "function", "inputs": [{"name": "_tokenAddress", "type": "address", "internalType": "address"}], "outputs": [], "stateMutability": "nonpayable"},
    {"name": "reservedToken", "type": "function", "inputs": [], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalClaimed", "type": "function", "inputs": [{"name": "_recipient", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalUnvested", "type": "function", "inputs": [{"name": "_recipient", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "totalVested", "type": "function", "inputs": [{"name": "_recipient", "type": "address", "internalType": "address"}], "outputs": [{"name": "", "type": "uint256", "internalType": "uint256"}], "stateMutability": "view"},
    {"name": "updateDisbursement", "type": "function", "inputs": [{"name": "_disbursementId", "type": "uint256", "internalType": "uint256"}, {"name": "_reduction", "type": "uint256", "internalType": "uint256"}, {"name": "_startTime", "type": "uint64", "internalType": "uint64"}, {"name": "_duration", "type": "uint64", "internalType": "uint64"}], "outputs": [], "stateMutability": "nonpayable"}
]
"""     

class TokenDisburse(ABIContractWrapper):
    def __init__(self, chain_key:str, rpc:str):
        contract_address = CONTRACT_ADDRESS[chain_key]
        super().__init__(contract_address=contract_address, abi=ABI, rpc=rpc)

    def add_disbursement(self, cred:Credentials, _recipient:address, _amount:uint256, _start_time:uint64, _duration:uint64, _note:string) -> TxReceipt:
        tx = self.contract.functions.addDisbursement(_recipient, _amount, _start_time, _duration, _note)
        return self.send_transaction(tx, cred)

    def add_disbursements(self, cred:Credentials, _recipients:Sequence[address], _amounts:Sequence[uint256], _start_time:uint64, _duration:uint64, _note:string) -> TxReceipt:
        tx = self.contract.functions.addDisbursements(_recipients, _amounts, _start_time, _duration, _note)
        return self.send_transaction(tx, cred)

    def claim(self, cred:Credentials, _amount:uint256) -> TxReceipt:
        tx = self.contract.functions.claim(_amount)
        return self.send_transaction(tx, cred)

    def disbursements(self, a:uint256, block_identifier:BlockIdentifier = 'latest') -> Tuple[uint256, address, string, uint256, uint256, uint256, uint64, uint64, uint64]:
        return self.contract.functions.disbursements(a).call(block_identifier=block_identifier)

    def get_disbursement_ids(self, _recipient:address, block_identifier:BlockIdentifier = 'latest') -> List[uint256]:
        return self.contract.functions.getDisbursementIds(_recipient).call(block_identifier=block_identifier)

    def initialize(self, cred:Credentials, _token_address:address) -> TxReceipt:
        tx = self.contract.functions.initialize(_token_address)
        return self.send_transaction(tx, cred)

    def reserved_token(self, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.reservedToken().call(block_identifier=block_identifier)

    def total_claimed(self, _recipient:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalClaimed(_recipient).call(block_identifier=block_identifier)

    def total_unvested(self, _recipient:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalUnvested(_recipient).call(block_identifier=block_identifier)

    def total_vested(self, _recipient:address, block_identifier:BlockIdentifier = 'latest') -> uint256:
        return self.contract.functions.totalVested(_recipient).call(block_identifier=block_identifier)

    def update_disbursement(self, cred:Credentials, _disbursement_id:uint256, _reduction:uint256, _start_time:uint64, _duration:uint64) -> TxReceipt:
        tx = self.contract.functions.updateDisbursement(_disbursement_id, _reduction, _start_time, _duration)
        return self.send_transaction(tx, cred)